"""
exponentially weighted moving average
"""

import numpy as np
import pandas as pd

# local module
import models_util

def ewm_fit_forecast(ts, fcst_len, span):
    
    try:
        
        ewm_model = pd.Series(ts).ewm(span=span)
        ewm_fittedvalues = ewm_model.mean()
        
        ewm_forecast = np.empty(fcst_len)
        ewm_forecast.fill(ewm_fittedvalues.iloc[-1])
        ewm_forecast = pd.Series(ewm_forecast)
        err = None

    except Exception as e:
        
        ewm_fittedvalues = None
        ewm_forecast = None
        err = str(e)
        
    return ewm_fittedvalues, ewm_forecast, err    

def fit_ewm(data_dict, freq, fcst_len, model_params, run_type, epsilon):
    
    model = 'ewm_model'
    ewm_params = model_params[model]
    span = 5
    
    complete_fact = data_dict['complete_fact']
    ewm_wfa = None
    
    ewm_fitted_values, ewm_forecast, err = ewm_fit_forecast(
                                                        ts = complete_fact['y'],
                                                        fcst_len = fcst_len,
                                                        span = span,

                                                    )
    
    
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        fitted_values, forecast, err = ewm_fit_forecast(
                                                            ts = train_fact['y'],
                                                            fcst_len = len(test_fact) ,
                                                            span = span
                                                        )
        
        if err is None:
            ewm_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = forecast.values,
                                    epsilon = epsilon,
                                )
            ewm_fitted_values = fitted_values.append(forecast)
            
        else:
            ewm_wfa = -1
            
    return ewm_fitted_values, ewm_forecast, ewm_wfa, err










# -*- coding: utf-8 -*-
"""
combine user input with parameters configuration and stat models configuration
"""

import os
import pandas as pd
import numpy as np
import datetime
import matplotlib.pyplot as plt

import utils
import compose

# default model params
model_params = utils.read_json('model_params.json')

# parameters configuration
param_config = utils.read_json('param_config.json')


"""
TODO: prompt user to accept default or set values
"""

"""
TODO: provide function documentation
"""

"""
run_type:
    all_models: return the result for all models without comparing out of sample forecast accuracy
    best_model: return the result for model with highest out of sample forecast accuracy
    all_best:   return the result for all models and indicate the model with highest out of sample forecast accuracy
"""

def collect_result(
                        raw_fact,
                        endog_colname,
                        gbkey,
                        ds_column, 
                        freq, 
                        fcst_length, 
                        run_type, 
                        holdout_length, 
                        model_list
                    ):
    
    
    pre_processed_dict, param_config = _pre_process_user_inputs(
                                                                    raw_fact,
                                                                    endog_colname,
                                                                    gbkey,
                                                                    ds_column, 
                                                                    freq, 
                                                                    fcst_length, 
                                                                    run_type, 
                                                                    holdout_length, 
                                                                    model_list
                                                            )
    
    result, fit_result_list = compose.compose_fit(pre_processed_dict, model_params, param_config, gbkey, run_type)
    
    return result, fit_result_list


def _pre_process_user_inputs(
                                    raw_fact,
                                    endog_colname,
                                    gbkey,
                                    ds_column, 
                                    freq, 
                                    fcst_length, 
                                    run_type, 
                                    holdout_length, 
                                    model_list
                            ):
    """
    add or rename date stamp column in input dataframe
    if comparing models results, separate input data to train-test dataframes
    """
    
    # next 4-lines can be done better
    
    param_config['FREQ'] = freq
    param_config['FORECAST_LEN'] = fcst_length
    param_config['HOLDOUT_LEN'] = holdout_length
    param_config['MODEL_LIST'] = model_list
    
    # replace endog column name with 'y'
    
    if len(raw_fact.columns) == 1:
        raw_fact.columns = ['y']
    else:
        if endog_colname in raw_fact_cols:
            raw_fact.rename(columns={endog_colname: 'y'}, inplace=True)
        else:
            #TODO: display this error to user
            raise ValueError('time series column name not found!!!')

    
    # add or rename date-time column
    
    if ds_column not in raw_fact.columns:
        raw_fact['date_stamp'] = pd.date_range(end=datetime.datetime.now(), periods=len(raw_fact), freq='D')
        
    else:
        raw_fact.rename(columns={ds_column: 'date_stamp'}, inplace=True)
        raw_fact['date_stamp'] = pd.to_datetime(raw_fact['date_stamp'])
        
    #TODO: if user chose to compare models, then create train-holdout sets
    #TODO: missing data interpolation, needs user input!
    
    pre_processed_dict = dict()    
        
    if run_type in ['best_model', 'all_best']:
        
        if gbkey in raw_fact.columns:
            for k,v in raw_fact.groupby(gbkey):
                train_fact = v.iloc[:-holdout_length]
                test_fact = v.iloc[-holdout_length:]

                pre_processed_dict[k] = {'complete_fact': v, 'train_fact': train_fact, 'test_fact': test_fact}
                
        else:
            train_fact = raw_fact.iloc[:-holdout_length]
            test_fact = raw_fact.iloc[-holdout_length:]
            pre_processed_dict[1] = {'complete_fact': raw_fact, 'train_fact': train_fact, 'test_fact': test_fact}
            
            
    else:
        if gbkey in raw_fact.columns:
            for k,v in raw_fact.groupby(gbkey):
                pre_processed_dict[k] = {'complete_fact': v}
                
        else:
            pre_processed_dict[1] = {'complete_fact': raw_fact}
            
    
    return pre_processed_dict, param_config
  






"""
UI built using dash to perform basic tasks like uploading user data and setting proper parameters.
"""


import base64
import datetime
import io
import os
import sys
import pandas as pd

import dash
from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_html_components as html
import dash_table

# import local modules

module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path+'\\scripts')

import main


external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

app = dash.Dash(__name__, external_stylesheets=external_stylesheets)

 
app.layout = html.Div([
    
    ### prompt user to provide endog column name ###
    
    dcc.Markdown('''
    **Provide time series values column name. Default is "y".**
    '''),
    dcc.Input(id="endog-column", type="text", placeholder="time series column", value='y'),
    html.Br(),   
    
    ### prompt user to provide id column name ###
    
    dcc.Markdown('''
    **If uploading more than one time series, provide an id column name. Default is "id".**
    '''),
    dcc.Input(id="id-column", type="text", placeholder="time series id column", value='id'),
    html.Br(),   
    
    ### prompt user to provide date_stamp column name ###
    
    dcc.Markdown('''
    **Provide date stamp column name if available. Default is "date_stamp".**
    '''),
    dcc.Input(id="ds-column", type="text", placeholder="Date-Time column", value='date_stamp'),
    html.Br(),   
    
    ### prompt user to provide time series frequency ###
    
    dcc.Markdown('''
    **Provide time series frequency. Default is 1.**
    '''),
    dcc.Input(id="ts-freq", type="number", placeholder='time series frequency', value=1),
    html.Br(),   
    
    ### prompt user to provide forecast length ###
    
    dcc.Markdown('''
    **Provide forecast length. Default is 10.**
    '''),
    dcc.Input(id="fcst-length", type="number", placeholder='forecast horizon', value=10),
    html.Br(),   
    
    ### display available output formats ###

    dcc.Markdown('''
    **Select result output format.**
    '''),
    dcc.RadioItems(
        id='result-format',
        options=[
            {'label': 'Best Model', 'value': 'best_model'},
            {'label': 'All Models', 'value': 'all_models'},
            {'label': 'All & Best', 'value': 'all_best'}
        ],
        value='all_models',
        labelStyle={'display': 'inline-block'}
    ),
    html.Br(),   
    
    ### prompt user to provide holdout length ###
    
    dcc.Markdown('''
    **If comparing model results, provide holdout length for out of sample
    forecast accuracy estimation. Default is 5.**
    '''),
    dcc.Input(id="holdout-length", type="number", placeholder='out of sample holdout len', value=5),
    html.Br(),   
    
    # display model list

    html.H5('Select or Remove Models'),
    
    dcc.Checklist(
        id='model-options',
        options=[
            {'label': 'EWM', 'value': 'ewm_model'},
            {'label': 'FFT', 'value': 'fft'},
            {'label': 'Holt Winters', 'value': 'holt_winters'},
            {'label': 'Prophet', 'value': 'prophet'},
            {'label': 'Sarimax', 'value': 'sarimax'}
        ],
        value=['ewm_model', 'fft', 'holt_winters', 'prophet', 'sarimax'],
        labelStyle={'display': 'inline-block'}
    ),
    
    html.Hr(),
    
    ### upload file box ###
    
    dcc.Upload(
        id='upload-data',
        children=html.Div([
            html.H6('Column name (header) is required!!!'),
            'Drag and Drop or ',
            html.A('Select Files'),
        ]),
        style={
            'width': '50%',
            'height': '100px',
            'lineHeight': '60px',
            'borderWidth': '1px',
            'borderStyle': 'dashed',
            'borderRadius': '5px',
            'textAlign': 'center',
            'margin': '10px'
        },
        multiple=True
    ),
    
    ### display a sample of input data ###
    
    html.Div(id='output-sample-data', style={'width': '50%'}),
    html.Br(),   
    
    ### display final result dataframe ###
    
    html.Div(
        id='output-result',
        style={'width': '50%'}
    ),
    html.Hr(),
    
])

### read user input
def read_contents(contents, ds_colname, filename):
    
    try:
        content_type, content_string = contents.split(',')
        decoded = base64.b64decode(content_string)
        
        if 'csv' in filename:
            # Assume that the user uploaded a CSV file
            df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))
            
            if ds_colname in df.columns:
                df[ds_colname] = pd.to_datetime(df[ds_colname])
            
            return df
            
        elif 'xls' in filename:
            # Assume that the user uploaded an excel file
            df = pd.read_excel(io.BytesIO(decoded))
            
            if ds_colname in df.columns:
                df[ds_colname] = pd.to_datetime(df[ds_colname])

            return df
        
        else:
            # TODO: update this to read txt file
            # Assume that the user uploaded a txt file
            df = pd.read_csv(
                io.StringIO(decoded.decode('utf-8'))
            )
            if ds_colname in df.columns:
                df[ds_colname] = pd.to_datetime(df[ds_colname])
            
            return df
        
    except Exception as e:
        return str(e)

### display user input
def parse_contents(contents, ds_colname, filename):
    
    content = read_contents(contents, ds_colname, filename)
    
    if type(content) == type(pd.DataFrame()):
        try:
            df_info = [col + ': ' + str(content[col].dtype) for col in content.columns]
            df_info = [x + ' *** ' for x in df_info]
            df_info = '\n'.join([s for s in df_info])

            return html.Div([
                #TODO: fix linebreak
                dcc.Markdown(df_info),

                html.H6('Input data: ' + filename),
                dash_table.DataTable(

                    data=content.to_dict('records'),
                    columns=[{'name': i, 'id': i} for i in content.columns],
                    page_action='none',
                    style_table={'height': '200px', 'overflowY': 'auto'},
                    
                ),
                html.Hr(),
            ])

        except Exception as e:
            return html.H6('Failed to display input data: ' + str(e))
        
        
    else:
        return html.H6('Failed to read input data: ' + content)

        
    

### display dataframe
def display_dataframe(df, name):
    
    if type(df) == type(pd.DataFrame()):
        return html.Div([
            html.H5(name),
            dash_table.DataTable(
                data=df.to_dict('records'),
                columns=[{'name': i, 'id': i} for i in df.columns],
                page_action='none',
                style_table={'height': '300px', 'overflowY': 'auto'},              
                export_columns = 'all',
                export_format = 'csv',
            ),
            html.Hr(),
        ])
    
    else:
        return html.Div(['run failed with error : ' + str(df)])
        


### read input file, display a sample
@app.callback(Output('output-sample-data', 'children'),
              [
                  Input('upload-data', 'contents'),
                  Input('ds-column', 'value'),
              ],
              [
                  State('upload-data', 'filename'),
              ])

def update_output(content_list, ds_colname, filename_list):
    
    if content_list is not None:
        return parse_contents(content_list[0], ds_colname, filename_list[0])        
    

### read input file and parameters, display forecast results
@app.callback(Output('output-result', 'children'),
              [
                  Input('endog-column', 'value'),
                  Input('id-column', 'value'),
                  Input('ds-column', 'value'),
                  Input('ts-freq', 'value'),
                  Input('fcst-length', 'value'),
                  Input('result-format', 'value'),
                  Input('holdout-length', 'value'),
                  Input('model-options', 'value'),
                  Input('upload-data', 'contents'),
              ],
              [State('upload-data', 'filename')])

def parse_result(endog_colname, gbkey, ds_colname, freq, fcst_length, run_type, holdout_length, model_list, content_list, filename_list):
    if content_list is not None:
        try:
            raw_fact = read_contents(content_list[0], ds_colname, filename_list[0])
            
            raw_fact_cols = raw_fact.columns
            
            #TODO: add flag for endog column
            
            if gbkey not in raw_fact_cols:
                gbkey = 'id'
            if ds_colname not in raw_fact_cols:
                ds_colname = 'date_stamp'
            if freq == '':
                freq = 1
            if fcst_length == '':
                fcst_length = 10
            if holdout_length == '':
                holdout_length = 5
                
            result, fit_result_list = main.collect_result(
                                                                 raw_fact,
                                                                 endog_colname,
                                                                 gbkey,
                                                                 ds_colname, 
                                                                 freq, 
                                                                 fcst_length, 
                                                                 run_type, 
                                                                 holdout_length, 
                                                                 model_list
                                                        )

            return display_dataframe(result, 'forecast result')
        
        except Exception as e:
            return display_dataframe(str(e), None)
    
    
    
    
if __name__ == '__main__':
    app.run_server(debug=True)


"""
collection of utility functions used in models section
"""

import numpy as np

def compute_wfa(y, yhat, epsilon):
    
    """
    weighted forecast accuracy
    """
    
    try:
        abs_err = compute_abs_err(y, yhat)
        wfa = max(0, 1-(2*sum(abs_err))/(sum(y) + sum(yhat) + epsilon))
        
    except Exception as e:
        raise ValueError('error call from wfa func: ' + str(e))
    
    return wfa

def compute_abs_err(y, yhat):
    
    #TODO: add more checks, e.g. data-type, same length, ...
    
    if any(np.isnan(y)) or any(np.isnan(yhat)):
        raise ValueError('error call from abs_err func: y or yhat has null values')
        
    if len(y) == 0 or len(yhat)  == 0:
        raise ValueError('error call from abs_err func: y or yhat is empty')
        
    return abs(y - yhat)






"""
Holt-Winters
"""

from statsmodels.tsa.holtwinters import ExponentialSmoothing

# local module
import models_util

def holt_winters_fit_forecast(ts, fcst_len, freq):
    
    try:                           
        hw_model = ExponentialSmoothing(
                                            endog = ts,
                                            trend = 'add',
                                            damped = True,
                                            seasonal = 'add',
                                            seasonal_periods = freq
                                          ).fit(
                                                optimized = True
                                               )

        hw_fittedvalues = hw_model.fittedvalues    

        hw_forecast = hw_model.predict(
                                            start = len(ts),
                                            end =   len(ts) + fcst_len - 1
                                        )
        err = None
        
    except Exception as e:
        hw_fittedvalues = None
        hw_forecast = None
        err = str(e)
    
    
    return hw_fittedvalues, hw_forecast, err

def fit_holt_winters(data_dict, freq, fcst_len, model_params, run_type, epsilon):
    
    model = 'holt_winters'
    hw_params = model_params[model]
    complete_fact = data_dict['complete_fact']
    hw_wfa = None
    
    
    hw_fitted_values, hw_forecast, err = holt_winters_fit_forecast(
                                                                       ts = complete_fact['y'],
                                                                       fcst_len = fcst_len,
                                                                       freq = freq,
                                                                  )
    
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']    
        fitted_values, forecast, err = holt_winters_fit_forecast(
                                                                           ts = train_fact['y'],
                                                                           fcst_len = len(test_fact),
                                                                           freq = freq,
                                                                 )
        
        if err is None:
            hw_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = forecast.values,
                                    epsilon = epsilon,
                                )
            hw_fitted_values = fitted_values.append(forecast)
            
        else:
            hw_wfa = -1
            
    return hw_fitted_values, hw_forecast, hw_wfa, err







# -*- coding: utf-8 -*-
"""
fitter class
"""

import os
import sys

# import local modules

module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path+'\\models')
    
from holt_winters import fit_holt_winters
from sarimax import fit_sarimax
from ewm import fit_ewm
from prophet import fit_prophet
from fft import fit_fft

class fitter:
    
    FIT_MODELS = {
                    'holt_winters':     fit_holt_winters,
                    'sarimax':          fit_sarimax,
                    'ewm_model':        fit_ewm,
                    'prophet':          fit_prophet,
                    'fft':              fit_fft,
                 }
    
    
    def __init__(self, model):
         self.model = model
         
    
    def fit(self, data_dict, freq, forecast_len, model_params, run_type, epsilon):
        """

        :param data_param_dict: ts values and parameters
        """
        fit_model = self.FIT_MODELS[self.model]
        
        return fit_model(data_dict, freq, forecast_len, model_params, run_type, epsilon)


# -*- coding: utf-8 -*-
"""
local utility functions 
"""

import os
import json
import pandas as pd
import numpy as np

def read_json(file_name):
    
    json_file = os.path.join('..\\params\\' + file_name)
    
    json_string = open(json_file).read()
    
    return json.loads(json_string.replace('\n', ' ').replace('\t', ' '))


def read_csv(file_name):
    
    csv_file = os.path.join('..\\data\\' + file_name)
    
    return pd.read_csv(csv_file)







"""
sarimax from statsmodels
"""

import statsmodels.api

# local module
import models_util

def sarimax_fit_forecast(ts, fcst_len):
    
    try:
        
        sarimax_model = statsmodels.api.tsa.statespace.SARIMAX(
                                                                    endog = ts,
                                                                    enforce_stationarity = False,
                                                                    enforce_invertibility = False,
                                                               ).fit()
        
        sarimax_fittedvalues = sarimax_model.fittedvalues
        sarimax_forecast = sarimax_model.predict(
                                                    start = len(ts),
                                                    end = len(ts) + fcst_len - 1,
                                                 )
        err = None
        
    except Exception as e:
        
        sarimax_model = None
        sarimax_fittedvalues = None
        sarimax_forecast = None
        err = str(e)
        
    return sarimax_fittedvalues, sarimax_forecast, err


def fit_sarimax(data_dict, freq, fcst_len, model_params, run_type, epsilon):
    
    model = 'sarimax'
    sarimax_params = model_params[model]
    
    complete_fact = data_dict['complete_fact']
    sarimax_wfa = None
    
    sarimax_fitted_values, sarimax_forecast, err = sarimax_fit_forecast(
                                                                            ts = complete_fact['y'],
                                                                            fcst_len = fcst_len,
                                                                        )
    
    
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        fitted_values, forecast, err = sarimax_fit_forecast(
                                                            ts = train_fact['y'],
                                                            fcst_len = len(test_fact) ,
                                                        )
        
        if err is None:
            sarimax_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = forecast.values,
                                    epsilon = epsilon,
                                )
            sarimax_fitted_values = fitted_values.append(forecast)
            
        else:
            sarimax_wfa = -1
            
    return sarimax_fitted_values, sarimax_forecast, sarimax_wfa, err





"""
Prophet: facebook's time series forecasting platform.
"""

from fbprophet import Prophet
import pandas as pd
import os

# local module
import models_util

class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    '''
    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)

            
            
def prophet_fit_forecast(df, fcst_len, freq):
    
    df = df[['date_stamp','y']].reset_index(drop=True)
    df.columns = ['ds', 'y']
    
    if freq == 12:
        prophet_freq = 'M'
    elif freq == 52:
        prophet_freq = 'W'
    else:
        prophet_freq = None
    
    try:
        
        prophet_model = Prophet()
        
        #with suppress_stdout_stderr():
        prophet_model.fit(df)
            
        future = prophet_model.make_future_dataframe(periods=fcst_len, freq=prophet_freq, include_history=True)
        
        prophet_model_predictions = prophet_model.predict(future)['yhat']
        
        prophet_fittedvalues = prophet_model_predictions.head(len(df))                
        
        prophet_forecast = prophet_model_predictions.tail(fcst_len)
        
        err = None
        
    except Exception as e:
        
        prophet_fittedvalues = None
        prophet_forecast = None
        err = str(e)
        
    return prophet_fittedvalues, prophet_forecast, err


def fit_prophet(data_dict, freq, fcst_len, model_params, run_type, epsilon):
    
    model = 'ewm_model'
    prophet_params = model_params[model]
    
    complete_fact = data_dict['complete_fact']
    prophet_wfa = None
    
    prophet_fitted_values, prophet_forecast, err = prophet_fit_forecast(
                                                                            df = complete_fact,
                                                                            fcst_len = fcst_len,
                                                                            freq = freq,
                                                                        )
    
    
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        fitted_values, forecast, err = prophet_fit_forecast(
                                                                df = train_fact,
                                                                fcst_len = len(test_fact) ,
                                                                freq = freq
                                                            )
        
        if err is None:
            prophet_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = forecast.values,
                                    epsilon = epsilon,
                                )
            prophet_fitted_values = fitted_values.append(forecast)
            
        else:
            prophet_wfa = -1
            
    return prophet_fitted_values, prophet_forecast, prophet_wfa, err




"""
fast fourier transform
"""

import numpy
import pandas

# local module
import models_util


def reconstruct_signal(
                         n_periods,
                         forecast_len,
                         fft_model,
                         ft_sample_frequencies,
                         fft_terms_for_reconstruction,
                         linear_trend
                      ):
    """

    :param n_periods:
    :param fft_model:
    :param ft_sample_frequencies:
    :param fft_terms_for_reconstruction:
    :param linear_trend:
    :return:
    """
    
    pi = numpy.pi
    t = numpy.arange(0, n_periods+forecast_len)
    restored_sig = numpy.zeros(t.size)
    for i in fft_terms_for_reconstruction:
        ampli = numpy.absolute(fft_model[i]) / n_periods
        phase = numpy.angle(
                             fft_model[i],
                             deg = False
                           )
        restored_sig += ampli * numpy.cos(2 * pi * ft_sample_frequencies[i] * t + phase)
    return restored_sig + linear_trend[0] * t


def fft_fit_forecast(ts, fcst_len, n_harmonics):
    
    ts_len = len(ts)
    
    try:
        t = numpy.arange(0, ts_len)
        linear_trend = numpy.polyfit(t, ts, 1)
        training_endog_detrend = ts - linear_trend[0] * t
        fft_model = numpy.fft.fft(training_endog_detrend)
        indexes = list(range(ts_len))
        
        # sort by amplitude
        indexes.sort(
                        key = lambda i: numpy.absolute(fft_model[i]) / ts_len,
                        reverse = True
                    )
        fft_terms_for_reconstruction = indexes[:1 + n_harmonics * 2]
        ft_sample_frequencies = numpy.fft.fftfreq(
                                                    n = ts_len,
                                                    d = 1
                                                 )
        
        fft_fit_forecast = reconstruct_signal(
                                                 n_periods = ts_len,
                                                 forecast_len = fcst_len,
                                                 fft_model = fft_model,
                                                 ft_sample_frequencies = ft_sample_frequencies,
                                                 fft_terms_for_reconstruction = fft_terms_for_reconstruction,
                                                 linear_trend = linear_trend
                                              )
        
        fft_fit_forecast = pandas.Series(fft_fit_forecast).clip(lower=0)
        
        
        fft_fittedvalues = fft_fit_forecast[:-(fcst_len)]
        
        fft_forecast = fft_fit_forecast[-(fcst_len):]
        
        err = None
        
        
    except Exception as e:
        fft_fittedvalues = None
        fft_forecast = None
        err = str(e)
        
    return fft_fittedvalues, fft_forecast, err


def fit_fft(data_dict, freq, fcst_len, model_params, run_type, epsilon):
    
    model = 'fft'
    fft_params = model_params[model]
    n_harmonics = 5
    
    complete_fact = data_dict['complete_fact']
    fft_wfa = None
    
    fft_fitted_values, fft_forecast, err = fft_fit_forecast(
                                                        ts = complete_fact['y'],
                                                        fcst_len = fcst_len,
                                                        n_harmonics = n_harmonics,

                                                    )
    
    
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        fitted_values, forecast, err = fft_fit_forecast(
                                                            ts = train_fact['y'],
                                                            fcst_len = len(test_fact),
                                                            n_harmonics = n_harmonics,
                                                        )
        
        if err is None:
            fft_wfa = models_util.compute_wfa(
                                                    y = test_fact['y'].values,
                                                    yhat = forecast.values,
                                                    epsilon = epsilon,
                                            )
            fft_fitted_values = fitted_values.append(forecast)
            
        else:
            fft_wfa = -1
            
    return fft_fitted_values, fft_forecast, fft_wfa, err



# -*- coding: utf-8 -*-
"""
manage input and output of forecasting models
"""

import pandas as pd
import numpy as np
import datetime
import fitter



def compose_fit(pre_processed_dict, model_params, param_config, gbkey, run_type):
    
    freq = param_config['FREQ']
    forecast_len = param_config['FORECAST_LEN']
    model_list = param_config['MODEL_LIST']
    epsilon = param_config['EPSILON']
    
    # list of dictionaries
    fit_result_list = list()
    
    for k, data_dict  in pre_processed_dict.items():
        
        fit_result = dict()
        
        fit_result['ts_id'] = k
        
        for m in model_list:

            f = fitter.fitter(m)

            (
             fit_result[m+'_fitted_values'],
             fit_result[m+'_forecast'],
             fit_result[m+'_wfa'],
             fit_result[m+'_err']
             ) = f.fit(data_dict, freq, forecast_len, model_params, run_type, epsilon)
            
        fit_result_list.append(fit_result)
        
    result = combine_to_dataframe(fit_result_list, model_list, run_type)
        
    return result, fit_result_list



def combine_to_dataframe(fit_result_list, model_list, run_type):
    
    #TODO: find the best model
    #TODO: return result

    if run_type == 'best_model':

        fit_result_list = _find_best_model(fit_result_list, model_list)
        
        df_list = list()

        for result_dict in fit_result_list:

            bm = result_dict['best_model']

            df = pd.DataFrame()

            try:
                fcst = result_dict[bm+'_fitted_values'].append(result_dict[bm+'_forecast']).values
                df[bm+'_forecast'] = fcst
                df[bm+'_wfa'] = result_dict[bm+'_wfa']

            except Exception as e:
                df[bm+'_forecast'] = 0
                df[bm+'_wfa'] = -1
                print(str(e))

            df['ts_id'] = result_dict['ts_id']
            df['best_model'] = result_dict['best_model']
            df_list.append(df)

        result = pd.concat(df_list, axis=0, ignore_index=True)
        
    elif run_type == 'all_best':
        
        fit_result_list = _find_best_model(fit_result_list, model_list)
        
        df_list = list()

        for result_dict in fit_result_list:
            df = pd.DataFrame()
            
            for m in model_list:

                try:
                    fcst = result_dict[m+'_fitted_values'].append(result_dict[m+'_forecast']).values
                    df[m+'_forecast'] = fcst
                    df[m+'_wfa'] = result_dict[m+'_wfa']

                except Exception as e:
                    df[m+'_forecast'] = 0
                    df[m+'_wfa'] = -1
                    print(str(e))

            df['ts_id'] = result_dict['ts_id']
            df['best_model'] = result_dict['best_model']
            df_list.append(df)

        result = pd.concat(df_list, axis=0, ignore_index=True)
                
        
    else:
        df_list = list()
        
        for result_dict in fit_result_list:
            df = pd.DataFrame()

            for m in model_list:

                try:
                    fcst = result_dict[m+'_fitted_values'].append(result_dict[m+'_forecast']).values
                    df[m+'_forecast'] = fcst

                except Exception as e:
                    df[m+'_forecast'] = 0
                    print(str(e))

            df['ts_id'] = result_dict['ts_id']
            df_list.append(df)
        
        result = pd.concat(df_list, axis=0, ignore_index=True) 
    
    return result


def _find_best_model(fit_result_list, model_list):
    
    for result_dict in fit_result_list:
        model_wfa_dict = dict()
        
        for m in model_list:
            model_wfa_dict[m] = result_dict[m+'_wfa']
            
        result_dict['best_model'] = [k for k,v in model_wfa_dict.items() if v == max(model_wfa_dict.values())][0]
    
    return fit_result_list


# not in use
def _transform_dataframe_to_dict(raw_fact, gbkey):
    
    data_param_list = list()
    
    if gbkey is None:
        
        data_param_list.append({'ts_id':1, 'df':raw_fact})
    
    else:
        for k,v in raw_fact.groupby(gbkey):
            data_param_list.append({'ts_id':k, 'df':v})
    
    return data_param_list





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
					name='foresee',
					version='0.1.0a4',
					author='Hamid Mohammadi',
					author_email='hmohammadi6545@gmail.com',
					description='Generate forecasts using several time series forecasting models in python.',
					long_description=long_description,
					long_description_content_type='text/markdown',
					url='https://github.com/HamidM6/foresee',
					packages=setuptools.find_packages(),
					classifiers=[
						"Programming Language :: Python :: 3",
						"License :: OSI Approved :: MIT License",
						"Operating System :: OS Independent",
					],
				)